import json
from lib.utils.sessions.session import Session
from lib.utils.sessions.sessions import Sessions

class Request():
    
    def __init__(self, data, sessions):
        """
            Permet de récupérer les informations d'une requête
            data: BaseHTTPRequestHandler, les informations de la requête
        """
        self.data = data
        self.sessions = sessions
        self.headers = data.headers
        self.path = data.path
        self.request_type = data.command
        self.server_address = data.client_address
        self.server_version = data.server_version
        self.protocol_version = data.protocol_version

        self.query = self._parse_query()
        self.body = self._parse_body()
        self.session = self._parse_session()

    def _parse_session(self):
        cookie = self.get('Cookie')
        if cookie:
            return self.sessions.get(cookie.split('=').pop())
        return Session(Sessions.random_session_id())

    def _parse_body(self):
        content_length = self.get('Content-Length')
        result = {}
        
        if content_length:
            body = self.data.rfile.read(int(content_length))
            if body:
                body = body.decode('utf8').replace("'", '"')
                result = json.loads(body)
        return result

    def _parse_query(self):
        chars = self.path.split('?')
        if len(chars) < 2:
            return {}
        chars = chars.pop().split('&')
        queries = {}
        for char in chars:
            char = char.split('=')
            queries[char[0]] = char[1]
        return queries

    def get(self, value):
        return self.headers.get(value)

    def accepts(self):
        return self.headers.get("accept")


from lib.utils.http_types import HttpMethods

class Route():

    def __init__(self, path, request_type, callback):
        self.path = path
        self.request_type = request_type
        self.callback = callback
    
    def get_method(self):
        return self.request_type

    def get_path(self):
        return self.path

    def handle(self, request, response):
        self.callback(request, response)

    def has_method(self, method):
        if not isinstance(method, HttpMethods):
            raise Exception("Please make sur your method is an instance of HttpMethods.")
        return method == self.get_method()


from http.server import BaseHTTPRequestHandler
from lib.utils.sessions.sessions import Sessions

import traceback

from .request import Request
from lib.utils.http_types import HttpMethods, HttpResponses
from .response import Response

sessions = Sessions()

class HttpRequestHandler(BaseHTTPRequestHandler):

    def __init__(self, routes, asset_directory, *args):
        global sessions

        self.routes = routes
        self.sessions = sessions
        self.asset_directory = asset_directory
        BaseHTTPRequestHandler.__init__(self, *args)

    def do_GET(self):
        result = self._is_asset()
        response = Response(self, self.sessions)

        if result:
            data, type, encoded = result
            return response.send(data, type, encoded)
        return self._handle_route(HttpMethods.GET, response)

    def do_POST(self):
        self._handle_route(HttpMethods.POST)

    def do_HEAD(self):
        self._handle_route(HttpMethods.HEAD)

    def do_PATCH(self):
        self._handle_route(HttpMethods.PATCH)

    def do_PUT(self):
        self._handle_route(HttpMethods.PUT)

    def do_DELETE(self):
        self._handle_route(HttpMethods.DELETE)
    
    def _handle_route(self, method, response = None):
        response = response or Response(self, self.sessions)
        path = self.path.split('?')[0]

        try: 
            for route in self.routes:
                if route['path'] == path:
                    route = route['route']
                    if route and (route.has_method(method) or route.has_method(HttpMethods.ALL)):
                        return route.handle(Request(self, self.sessions), response)
            return response.status(HttpResponses.NOT_FOUND) \
                .send(f'Cannot {method.value} {path}')
        except Exception:
            response.send(traceback.format_exc())

    def _is_asset(self):
        if self.asset_directory:
            if self.asset_directory['name'] in self.path:
                path = self.path.replace(self.asset_directory['name'], '')
                path = path.replace('\\', '/')
                path = self.asset_directory['directory'] + path
                type = path.split('.').pop()
                
                if type in ['png', 'jpg', 'gif', 'jfif', 'wepb', 'svg']:
                    return (bytearray(open(path, 'rb').read()), f'image/{type}', False)
                return (open(path, 'r').read(), f"text/{type}", True)
            return False
        return False


from http.server import HTTPServer

from .utils.http_types import HttpMethods
from .handlers.http_request_handler import HttpRequestHandler
from .utils.route import Route

class Application():

    def __init__(self):
        self.instance = None
        self.routes = list()
        self.asset_directory = {}
        pass

    def all(self, path, callback):
        self._add_route(path, HttpMethods.ALL, callback)

    def post(self, path, callback):
        self._add_route(path, HttpMethods.POST, callback)

    def get(self, path , callback):
        self._add_route(path, HttpMethods.GET, callback)

    def put(self, path, callback):
        self._add_route(path, HttpMethods.PUT, callback)

    def patch(self, path, callback):
        self._add_route(path, HttpMethods.PATCH, callback)
        
    def delete(self, path, callback):
        self._add_route(path, HttpMethods.DELETE, callback)

    def head(self, path, callback):
        self._add_route(path, HttpMethods.HEAD, callback)

    def define_asset(self, name, directory):
        self.asset_directory = {
            "name": name,
            "directory": directory
        }

    def listen(self, port, callback):
        if not callback or not hasattr(callback, '__call__'):
            raise ValueError("Callback must be a function.")
        if not port and type(port) is not int:
            callback(ValueError("No port found."))
        try:
            callback(None)
            self.instance = HTTPServer(('localhost', port), self._get_handler)
            self.instance.serve_forever()
        except:
            callback(ConnectionError("Impossible connection"))

    def _add_route(self, path, request_type, callback):
        if not path or type(path) is not str:
            raise ValueError("Path not found.")
        if not callback or not hasattr(callback, '__call__'):
            raise ValueError("Callback must be a function.")
        self.routes.append({
            "path": path,
            "route": Route(path, request_type, callback)
        })

    def _get_handler(self, *args):
        HttpRequestHandler(self.routes, self.asset_directory, *args)


from enum import Enum

class HttpResponses(Enum):
    UNAUTHORIZED = 401
    FORBIDDEN = 403
    NOT_FOUND = 404
    BAD_REQUEST = 400

    OK = 200
    NO_CONTENT = 204

    INTERNAL_SERVER_ERROR = 500

class HttpMethods(Enum):
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    HEAD = "HEAD"
    PATCH = "PATCH"
    ALL = "ALL"


import json
from os.path import exists

from lib.utils.http_types import HttpResponses 

class Response():

    def __init__(self, data, sessions):
        self.data = data
        self.headers = {}
        self.sessions = sessions
        self.currentStatus = HttpResponses.OK
        self.statusHandled = False
        

    def status(self, code):
        if not isinstance(code, HttpResponses):
            raise Exception("Please, make sur this is an instance of HTTPResponses")
        self.currentStatus = code
        if not self.statusHandled:
            self.statusHandled = True
            self.data.send_response_only(self.currentStatus.value)
        return self

    def end(self):
        self.data.flush_headers()
        return self

    def set_header(self, key, value):
        self.headers[key] = value
        return self

    def send_header(self, key, value):
        self.data.send_header(key, value)
        return self
        
    def send(self, data, type = None, encoded = True):
        self.status(self.currentStatus) \
            .set_header("Content-type", (type, "text/plain")[not type])
        self._store_session()
        if self.headers:
            for key, value in self.headers.items():
                self.send_header(key, value)
        self.data.end_headers()
        self.data.wfile.write(data.encode("utf_8") if encoded else data)
        return self

    def _store_session(self):
        cookie = self.data.headers.get('Cookie')

        if (not cookie) or (cookie and not self.sessions.has(cookie.split('=').pop())):
            session = self.sessions.add()
            self.set_header("Set-Cookie", session)
    
    def json(self, data):
        return self.send(json.dumps(data), "application/json")

    def render(self, path):
        if not path.startswith('./'):
            path = "./" + path
        if not path.endswith(".html"):
            path += ".html"
        if not exists(path):
            raise FileNotFoundError("HTML file not found.")
        file = open(path, 'r', encoding="utf-8")
        return self.send(file.read().replace('\n', ''), "text/html")


from setuptools import setup

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    name="asita",
    version="0.1.0",
    author="Mattéo Gaillard",
    description="A web application framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Matteo0810/PyServ",
    packages=[
        "asita",
        "asita.handlers",
        "asita.utils"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
)